@@ -1,24 +1,15 @@
-# -*- coding: utf-8 -*-
 import re
 
+from django.urls import NoReverseMatch, reverse
 from django.utils.safestring import mark_safe
 
-from crispy_forms.compatibility import string_types
 from crispy_forms.exceptions import FormHelpersException
 from crispy_forms.layout import Layout
 from crispy_forms.layout_slice import LayoutSlice
-from crispy_forms.utils import (
-    TEMPLATE_PACK, flatatt, list_difference, list_intersection, render_field,
-)
-
-try:
-    from django.urls import reverse, NoReverseMatch
-except ImportError:
-    # Django < 1.10
-    from django.core.urlresolvers import reverse, NoReverseMatch
+from crispy_forms.utils import TEMPLATE_PACK, flatatt, list_difference, render_field
 
 
-class DynamicLayoutHandler(object):
+class DynamicLayoutHandler:
     def _check_layout(self):
         if self.layout is None:
             raise FormHelpersException("You need to set a layout in your FormHelper")
@@ -40,8 +31,8 @@ class DynamicLayoutHandler(object):
         Returns a LayoutSlice pointing to layout objects of type `LayoutClass`
         """
         self._check_layout()
-        max_level = kwargs.pop('max_level', 0)
-        greedy = kwargs.pop('greedy', False)
+        max_level = kwargs.pop("max_level", 0)
+        greedy = kwargs.pop("greedy", False)
         filtered_layout_objects = self.layout.get_layout_objects(LayoutClasses, max_level=max_level, greedy=greedy)
 
         return LayoutSlice(self.layout, filtered_layout_objects)
@@ -82,7 +73,7 @@ class DynamicLayoutHandler(object):
         and not a copy.
         """
         # when key is a string containing the field name
-        if isinstance(key, string_types):
+        if isinstance(key, str):
             # Django templates access FormHelper attributes using dictionary [] operator
             # This could be a helper['form_id'] access, not looking for a field
             if hasattr(self, key):
@@ -125,13 +116,13 @@ class FormHelper(DynamicLayoutHandler):
     Let's see what attributes you can set and what form behaviors they apply to:
 
         **form_method**: Specifies form method attribute.
-            You can see it to 'POST' or 'GET'. Defaults to 'POST'
+            You can set it to 'POST' or 'GET'. Defaults to 'POST'
 
         **form_action**: Applied to the form action attribute:
             - Can be a named url in your URLconf that can be executed via the `{% url %}` template tag. \
             Example: 'show_my_profile'. In your URLconf you could have something like::
 
-                url(r'^show/profile/$', 'show_my_profile_view', name = 'show_my_profile')
+                path('show/profile/', 'show_my_profile_view', name = 'show_my_profile')
 
             - It can simply point to a URL '/whatever/blabla/'.
 
@@ -193,13 +184,14 @@ class FormHelper(DynamicLayoutHandler):
         {% load crispy_forms_tags %}
         {% crispy form %}
     """
-    _form_method = 'post'
-    _form_action = ''
-    _form_style = 'default'
+
+    _form_method = "post"
+    _form_action = ""
+    _form_style = "default"
     form = None
-    form_id = ''
-    form_class = ''
-    form_group_wrapper_class = ''
+    form_id = ""
+    form_class = ""
+    form_group_wrapper_class = ""
     layout = None
     form_tag = True
     form_error_title = None
@@ -215,8 +207,9 @@ class FormHelper(DynamicLayoutHandler):
     template = None
     field_template = None
     disable_csrf = False
-    label_class = ''
-    field_class = ''
+    use_custom_control = True
+    label_class = ""
+    field_class = ""
     include_media = True
 
     def __init__(self, form=None):
@@ -236,9 +229,11 @@ class FormHelper(DynamicLayoutHandler):
 
     @form_method.setter
     def form_method(self, method):
-        if method.lower() not in ('get', 'post'):
-            raise FormHelpersException('Only GET and POST are valid in the \
-                    form_method helper attribute')
+        if method.lower() not in ("get", "post"):
+            raise FormHelpersException(
+                "Only GET and POST are valid in the \
+                    form_method helper attribute"
+            )
 
         self._form_method = method.lower()
 
@@ -256,16 +251,18 @@ class FormHelper(DynamicLayoutHandler):
     @property
     def form_style(self):
         if self._form_style == "default":
-            return ''
+            return ""
 
         if self._form_style == "inline":
-            return 'inlineLabels'
+            return "inlineLabels"
 
     @form_style.setter
     def form_style(self, style):
-        if style.lower() not in ('default', 'inline'):
-            raise FormHelpersException('Only default and inline are valid in the \
-                    form_style helper attribute')
+        if style.lower() not in ("default", "inline"):
+            raise FormHelpersException(
+                "Only default and inline are valid in the \
+                    form_style helper attribute"
+            )
 
         self._form_style = style.lower()
 
@@ -301,111 +298,91 @@ class FormHelper(DynamicLayoutHandler):
         form.crispy_field_template = self.field_template
 
         # This renders the specified Layout strictly
-        html = self.layout.render(
-            form,
-            self.form_style,
-            context,
-            template_pack=template_pack
-        )
+        html = self.layout.render(form, self.form_style, context, template_pack=template_pack)
 
         # Rendering some extra fields if specified
         if self.render_unmentioned_fields or self.render_hidden_fields or self.render_required_fields:
-            fields = set(form.fields.keys())
-            left_fields_to_render = fields - form.rendered_fields
+            fields = tuple(form.fields.keys())
+            left_fields_to_render = list_difference(fields, form.rendered_fields)
             for field in left_fields_to_render:
                 if (
-                    self.render_unmentioned_fields or
-                    self.render_hidden_fields and form.fields[field].widget.is_hidden or
-                    self.render_required_fields and form.fields[field].widget.is_required
+                    self.render_unmentioned_fields
+                    or (self.render_hidden_fields and form.fields[field].widget.is_hidden)
+                    or (self.render_required_fields and form.fields[field].widget.is_required)
                 ):
-                    html += render_field(
-                        field,
-                        form,
-                        self.form_style,
-                        context,
-                        template_pack=template_pack
-                    )
-
-        # If the user has Meta.fields defined, not included in the layout,
-        # we suppose they need to be rendered
-        if hasattr(form, 'Meta'):
-            if hasattr(form.Meta, 'fields'):
-                current_fields = tuple(getattr(form, 'fields', {}).keys())
-                meta_fields = getattr(form.Meta, 'fields')
-
-                fields_to_render = list_intersection(current_fields, meta_fields)
-                left_fields_to_render = list_difference(fields_to_render, form.rendered_fields)
-
-                for field in left_fields_to_render:
-                    # We still respect the configuration of the helper
-                    # regarding which fields to render
-                    if (
-                        self.render_unmentioned_fields or
-                        (self.render_hidden_fields and
-                         form.fields[field].widget.is_hidden) or
-                        (self.render_required_fields and
-                         form.fields[field].widget.is_required)
-                    ):
-                        html += render_field(field, form, self.form_style, context)
+                    html += render_field(field, form, self.form_style, context, template_pack=template_pack)
 
         return mark_safe(html)
 
-    def get_attributes(self, template_pack=TEMPLATE_PACK):
+    def get_attributes(self, template_pack=TEMPLATE_PACK):  # noqa: C901
         """
         Used by crispy_forms_tags to get helper attributes
         """
         items = {
-            'form_method': self.form_method.strip(),
-            'form_tag': self.form_tag,
-            'form_style': self.form_style.strip(),
-            'form_show_errors': self.form_show_errors,
-            'help_text_inline': self.help_text_inline,
-            'error_text_inline': self.error_text_inline,
-            'html5_required': self.html5_required,
-            'form_show_labels': self.form_show_labels,
-            'disable_csrf': self.disable_csrf,
-            'label_class': self.label_class,
-            'field_class': self.field_class,
-            'include_media': self.include_media
+            "disable_csrf": self.disable_csrf,
+            "error_text_inline": self.error_text_inline,
+            "field_class": self.field_class,
+            "field_template": self.field_template or "%s/field.html" % template_pack,
+            "form_method": self.form_method.strip(),
+            "form_show_errors": self.form_show_errors,
+            "form_show_labels": self.form_show_labels,
+            "form_style": self.form_style.strip(),
+            "form_tag": self.form_tag,
+            "help_text_inline": self.help_text_inline,
+            "html5_required": self.html5_required,
+            "include_media": self.include_media,
+            "label_class": self.label_class,
+            "use_custom_control": self.use_custom_control,
         }
-        bootstrap_size_match = re.findall('col-(lg|md|sm|xs)-(\d+)', self.label_class)
-        if bootstrap_size_match:
-            if template_pack == 'bootstrap4':
-                offset_pattern = 'offset-%s-%s'
-            else:
-                offset_pattern = 'col-%s-offset-%s'
-            items['bootstrap_checkbox_offsets'] = [offset_pattern % m for m in bootstrap_size_match]
 
-        items['attrs'] = {}
+        if template_pack == "bootstrap4":
+            if "form-horizontal" in self.form_class.split():
+                bootstrap_size_match = re.findall(r"col(-(xl|lg|md|sm))?-(\d+)", self.label_class)
+                if bootstrap_size_match:
+                    offset_pattern = "offset%s-%s"
+                    items["bootstrap_checkbox_offsets"] = [
+                        offset_pattern % (m[0], m[-1]) for m in bootstrap_size_match
+                    ]
+        else:
+            bootstrap_size_match = re.findall(r"col-(lg|md|sm|xs)-(\d+)", self.label_class)
+            if bootstrap_size_match:
+                offset_pattern = "col-%s-offset-%s"
+                items["bootstrap_checkbox_offsets"] = [offset_pattern % m for m in bootstrap_size_match]
+
+        items["attrs"] = {}
         if self.attrs:
-            items['attrs'] = self.attrs.copy()
+            items["attrs"] = self.attrs.copy()
         if self.form_action:
-            items['attrs']['action'] = self.form_action.strip()
+            items["attrs"]["action"] = self.form_action.strip()
         if self.form_id:
-            items['attrs']['id'] = self.form_id.strip()
+            items["attrs"]["id"] = self.form_id.strip()
         if self.form_class:
             # uni_form TEMPLATE PACK has a uniForm class by default
-            if template_pack == 'uni_form':
-                items['attrs']['class'] = "uniForm %s" % self.form_class.strip()
+            if template_pack == "uni_form":
+                items["attrs"]["class"] = "uniForm %s" % self.form_class.strip()
             else:
-                items['attrs']['class'] = self.form_class.strip()
+                items["attrs"]["class"] = self.form_class.strip()
         else:
-            if template_pack == 'uni_form':
-                items['attrs']['class'] = self.attrs.get('class', '') + " uniForm"
+            if template_pack == "uni_form":
+                items["attrs"]["class"] = self.attrs.get("class", "") + " uniForm"
         if self.form_group_wrapper_class:
-            items['attrs']['form_group_wrapper_class'] = self.form_group_wrapper_class
+            items["attrs"]["form_group_wrapper_class"] = self.form_group_wrapper_class
 
-        items['flat_attrs'] = flatatt(items['attrs'])
+        items["flat_attrs"] = flatatt(items["attrs"])
 
         if self.inputs:
-            items['inputs'] = self.inputs
+            items["inputs"] = self.inputs
         if self.form_error_title:
-            items['form_error_title'] = self.form_error_title.strip()
+            items["form_error_title"] = self.form_error_title.strip()
         if self.formset_error_title:
-            items['formset_error_title'] = self.formset_error_title.strip()
+            items["formset_error_title"] = self.formset_error_title.strip()
 
         for attribute_name, value in self.__dict__.items():
-            if attribute_name not in items and attribute_name not in ['layout', 'inputs'] and not attribute_name.startswith('_'):
+            if (
+                attribute_name not in items
+                and attribute_name not in ["layout", "inputs"]
+                and not attribute_name.startswith("_")
+            ):
                 items[attribute_name] = value
 
         return items
