@@ -15,6 +15,7 @@ from crispy_forms.utils import render_crispy_form
 
 from .conftest import only_bootstrap, only_bootstrap3, only_bootstrap4, only_uni_form
 from .forms import (
+    AdvancedFileForm,
     CheckboxesSampleForm,
     CrispyEmptyChoiceTestModel,
     CrispyTestModel,
@@ -23,9 +24,11 @@ from .forms import (
     SampleForm2,
     SampleForm3,
     SampleForm4,
+    SampleForm5,
     SampleForm6,
+    SelectSampleForm,
 )
-from .utils import contains_partial
+from .utils import contains_partial, parse_expected, parse_form
 
 
 def test_invalid_unicode_characters(settings):
@@ -68,7 +71,7 @@ def test_meta_extra_fields_with_missing_fields():
     del form.fields["email"]
 
     form_helper = FormHelper()
-    form_helper.layout = Layout("first_name",)
+    form_helper.layout = Layout("first_name")
 
     template = Template(
         """
@@ -99,7 +102,7 @@ def test_layout_unresolved_field(settings):
 
 def test_double_rendered_field(settings):
     form_helper = FormHelper()
-    form_helper.add_layout(Layout("is_company", "is_company",))
+    form_helper.add_layout(Layout("is_company", "is_company"))
 
     template = Template(
         """
@@ -150,7 +153,12 @@ def test_layout_fieldset_row_html_with_unicode_fieldnames(settings):
             Fieldset(
                 "User Data",
                 "email",
-                Row("password1", "password2", css_id="row_passwords", css_class="rows",),
+                Row(
+                    "password1",
+                    "password2",
+                    css_id="row_passwords",
+                    css_class="rows",
+                ),
                 HTML('<a href="#" id="testLink">test link</a>'),
                 HTML(
                     """
@@ -201,8 +209,19 @@ def test_change_layout_dynamically_delete_field():
     form_helper = FormHelper()
     form_helper.add_layout(
         Layout(
-            Fieldset("Company Data", "is_company", "email", "password1", "password2", css_id="multifield_info",),
-            Column("first_name", "last_name", css_id="column_name",),
+            Fieldset(
+                "Company Data",
+                "is_company",
+                "email",
+                "password1",
+                "password2",
+                css_id="multifield_info",
+            ),
+            Column(
+                "first_name",
+                "last_name",
+                css_id="column_name",
+            ),
         )
     )
 
@@ -228,8 +247,15 @@ def test_column_has_css_classes(settings):
     form_helper = FormHelper()
     form_helper.add_layout(
         Layout(
-            Fieldset("Company Data", "is_company", "email", "password1", "password2", css_id="multifield_info",),
-            Column("first_name", "last_name",),
+            Fieldset(
+                "Company Data",
+                "is_company",
+                "email",
+                "password1",
+                "password2",
+                css_id="multifield_info",
+            ),
+            Column("first_name", "last_name"),
         )
     )
 
@@ -257,7 +283,7 @@ def test_bs4_column_css_classes(settings):
     form_helper = FormHelper()
     form_helper.add_layout(
         Layout(
-            Column("first_name", "last_name",),
+            Column("first_name", "last_name"),
             Column("first_name", "last_name", css_class="col-sm"),
             Column("first_name", "last_name", css_class="mb-4"),
         )
@@ -279,7 +305,11 @@ def test_formset_layout(settings):
     helper.form_method = "POST"
     helper.form_action = "simpleAction"
     helper.layout = Layout(
-        Fieldset("Item {{ forloop.counter }}", "is_company", "email",),
+        Fieldset(
+            "Item {{ forloop.counter }}",
+            "is_company",
+            "email",
+        ),
         HTML("{% if forloop.first %}Note for first form only{% endif %}"),
         Row("password1", "password2"),
         Fieldset("", "first_name", "last_name"),
@@ -358,7 +388,14 @@ def test_i18n():
     )
     form = SampleForm()
     form_helper = FormHelper()
-    form_helper.layout = Layout(HTML(_("i18n text")), Fieldset(_("i18n legend"), "first_name", "last_name",))
+    form_helper.layout = Layout(
+        HTML(_("i18n text")),
+        Fieldset(
+            _("i18n legend"),
+            "first_name",
+            "last_name",
+        ),
+    )
     form.helper = form_helper
 
     html = template.render(Context({"form": form}))
@@ -417,15 +454,29 @@ def test_layout_composition():
     form_helper = FormHelper()
     form_helper.add_layout(
         Layout(
-            Layout(MultiField("Some company data", "is_company", "email", css_id="multifield_info",),),
+            Layout(
+                MultiField(
+                    "Some company data",
+                    "is_company",
+                    "email",
+                    css_id="multifield_info",
+                ),
+            ),
             Column(
                 "first_name",
                 # 'last_name', Missing a field on purpose
                 css_id="column_name",
                 css_class="columns",
             ),
-            ButtonHolder(Submit("Save", "Save", css_class="button white"),),
-            Div("password1", "password2", css_id="custom-div", css_class="customdivs",),
+            ButtonHolder(
+                Submit("Save", "Save", css_class="button white"),
+            ),
+            Div(
+                "password1",
+                "password2",
+                css_id="custom-div",
+                css_class="customdivs",
+            ),
         )
     )
 
@@ -464,7 +515,12 @@ def test_second_layout_multifield_column_buttonholder_submit_div():
                 title="multifield_title",
                 multifield_test="123",
             ),
-            Column("first_name", "last_name", css_id="column_name", css_class="columns",),
+            Column(
+                "first_name",
+                "last_name",
+                css_id="column_name",
+                css_class="columns",
+            ),
             ButtonHolder(
                 Submit(
                     "Save the world", "{{ value_var }}", css_class="button white", data_id="test", data_name="test"
@@ -527,30 +583,90 @@ def test_keepcontext_context_manager(settings):
 
 
 @only_bootstrap4
-def test_use_custom_control_is_used():
+def test_use_custom_control_is_used_in_checkboxes():
     form = CheckboxesSampleForm()
     form.helper = FormHelper()
-    form.helper.layout = Layout("checkboxes", InlineCheckboxes("alphacheckboxes"), "numeric_multiple_checkboxes")
+    form.helper.layout = Layout(
+        "checkboxes",
+        InlineCheckboxes("alphacheckboxes"),
+        "numeric_multiple_checkboxes",
+    )
     # form.helper.use_custom_control take default value which is True
-
-    response = render(request=None, template_name="crispy_render_template.html", context={"form": form})
-    assert response.content.count(b"custom-control-inline") == 3
-    assert response.content.count(b"custom-checkbox") == 9
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_checkboxes_true.html"
+    )
 
     form.helper.use_custom_control = True
-
-    response = render(request=None, template_name="crispy_render_template.html", context={"form": form})
-    assert response.content.count(b"custom-control-inline") == 3
-    assert response.content.count(b"custom-checkbox") == 9
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_checkboxes_true.html"
+    )
 
     form.helper.use_custom_control = False
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_checkboxes_false.html"
+    )
 
-    response = render(request=None, template_name="crispy_render_template.html", context={"form": form})
+    form = CheckboxesSampleForm({})
+    form.helper = FormHelper()
+    form.helper.layout = Layout(
+        "checkboxes",
+        InlineCheckboxes("alphacheckboxes"),
+        "numeric_multiple_checkboxes",
+    )
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_checkboxes_true_failing.html"
+    )
 
-    assert response.content.count(b"custom-control-inline") == 0
-    assert response.content.count(b"form-check-inline") == 3
-    assert response.content.count(b"form-check") > 0
-    assert response.content.count(b"custom-checkbox") == 0
+
+@only_bootstrap4
+def test_use_custom_control_is_used_in_radio():
+    form = SampleForm5()
+    form.helper = FormHelper()
+    form.helper.layout = Layout(
+        "radio_select",
+    )
+    # form.helper.use_custom_control take default value which is True
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_radio_true.html"
+    )
+
+    form.helper.use_custom_control = True
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_radio_true.html"
+    )
+
+    form.helper.use_custom_control = False
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_radio_false.html"
+    )
+
+    form = SampleForm5({})
+    form.helper = FormHelper()
+    form.helper.layout = Layout(
+        "radio_select",
+    )
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_use_custom_control_is_used_in_radio_true_failing.html"
+    )
+
+
+@only_bootstrap4
+@pytest.mark.parametrize(
+    "use_custom_control, expected_html",
+    [
+        (True, "bootstrap4/test_layout/test_use_custom_control_in_select_true.html"),
+        (False, "bootstrap4/test_layout/test_use_custom_control_in_select_false.html"),
+    ],
+)
+def test_use_custom_control_in_select(use_custom_control, expected_html):
+    form = SelectSampleForm()
+
+    form.helper = FormHelper()
+    form.helper.template_pack = "bootstrap4"
+    form.helper.layout = Layout("select")
+    form.helper.use_custom_control = use_custom_control
+
+    assert parse_form(form) == parse_expected(expected_html)
 
 
 @only_bootstrap3
@@ -559,7 +675,11 @@ def test_form_inline():
     form.helper = FormHelper()
     form.helper.form_class = "form-inline"
     form.helper.field_template = "bootstrap3/layout/inline_field.html"
-    form.helper.layout = Layout("email", "password1", "last_name",)
+    form.helper.layout = Layout(
+        "email",
+        "password1",
+        "last_name",
+    )
 
     html = render_crispy_form(form)
     assert html.count('class="form-inline"') == 1
@@ -576,7 +696,7 @@ def test_bootstrap4_form_inline():
     form.helper = FormHelper()
     form.helper.form_class = "form-inline"
     form.helper.field_template = "bootstrap4/layout/inline_field.html"
-    form.helper.layout = Layout("email", "password1", "last_name",)
+    form.helper.layout = Layout("email", "password1", "last_name")
 
     html = render_crispy_form(form)
     assert html.count('class="form-inline"') == 1
@@ -590,12 +710,16 @@ def test_bootstrap4_form_inline():
 def test_update_attributes_class():
     form = SampleForm()
     form.helper = FormHelper()
-    form.helper.layout = Layout("email", Field("password1"), "password2",)
+    form.helper.layout = Layout("email", Field("password1"), "password2")
     form.helper["password1"].update_attributes(css_class="hello")
     html = render_crispy_form(form)
     assert html.count(' class="hello textinput') == 1
     form.helper = FormHelper()
-    form.helper.layout = Layout("email", Field("password1", css_class="hello"), "password2",)
+    form.helper.layout = Layout(
+        "email",
+        Field("password1", css_class="hello"),
+        "password2",
+    )
     form.helper["password1"].update_attributes(css_class="hello2")
     html = render_crispy_form(form)
     assert html.count(' class="hello hello2 textinput') == 1
@@ -606,25 +730,36 @@ def test_file_field():
     form = FileForm()
     form.helper = FormHelper()
     form.helper.layout = Layout("clearable_file")
-    html = render_crispy_form(form)
-    assert '<span class="custom-control custom-checkbox">' in html
-    assert '<input type="file" name="clearable_file" class="custom-file-input"  >' in html
+    assert parse_form(form) == parse_expected("bootstrap4/test_layout/test_file_field_clearable_custom_control.html")
 
     form.helper.use_custom_control = False
-    html = render_crispy_form(form)
-    assert '<input type="checkbox" name="clearable_file-clear" id="clearable_file-clear_id">' in html
-    assert '<input type="file" name="clearable_file" class="custom-file-input"  >' not in html
+    assert parse_form(form) == parse_expected("bootstrap4/test_layout/test_file_field_clearable.html")
 
     form.helper.use_custom_control = True
     form.helper.layout = Layout("file_field")
-    html = render_crispy_form(form)
-    assert '<div class="form-control custom-file"' in html
-    assert '<input type="file" name="file_field" class="custom-file-input"' in html
-    assert '<label class="custom-file-label' in html
-    assert 'for="id_file_field">---</label>' in html
+    assert parse_form(form) == parse_expected("bootstrap4/test_layout/test_file_field_custom_control.html")
 
     form.helper.use_custom_control = False
-    html = render_crispy_form(form)
-    assert "custom-file" not in html
-    assert "custom-file-input" not in html
-    assert "custom-file-label" not in html
+    assert parse_form(form) == parse_expected("bootstrap4/test_layout/test_file_field_default.html")
+
+
+@only_bootstrap4
+def test_file_field_with_custom_class():
+    form = AdvancedFileForm()
+    form.helper = FormHelper()
+    form.helper.layout = Layout("clearable_file")
+    assert parse_form(form) == parse_expected(
+        "bootstrap4/test_layout/test_file_field_with_custom_class_clearable.html"
+    )
+
+    form.helper.layout = Layout("file_field")
+    assert parse_form(form) == parse_expected("bootstrap4/test_layout/test_file_field_with_custom_class.html")
+
+
+@only_bootstrap4
+def test_form_control_size():
+    "CSS classes form-control and form-control-lg are both required"
+    form = SampleForm()
+    form.helper = FormHelper()
+    form.helper.layout = Layout(Field("first_name", css_class="form-control-lg"))
+    assert parse_form(form) == parse_expected("bootstrap4/test_layout/test_form_control_size.html")
